import pytest

from django.db import models
from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from rest_framework import mixins
from rest_framework import routers
from rest_framework import viewsets
from rest_framework_json_api import parsers
from rest_framework_json_api import relations
from rest_framework_json_api import renderers
from rest_framework_json_api import serializers

from tests import base
from tests import compatibility


class MemberWithCustomID(models.Model):
    custom_id = models.BigIntegerField(primary_key=True)
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)

    class Meta:
        app_label = 'tests'


class ProjectWithCustomID(models.Model):
    custom_id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=100)
    archived = models.BooleanField()
    members = models.ManyToManyField(MemberWithCustomID, related_name='projects')
    owner_member = models.ForeignKey(MemberWithCustomID, on_delete=models.DO_NOTHING, related_name='owned_projects')

    class Meta:
        app_label = 'tests'


@pytest.mark.parametrize(
    'read_only', (
        True,
        False,
    )
)
def test_related_resource(read_only):
    """
    Correctly select id from non default pk field for both model and related models
    """
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = ProjectWithCustomID
            fields = ('custom_id', 'name', 'archived', 'members', 'owner_member')
            read_only_fields = ['members', 'owner_member'] if read_only else []

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = ProjectWithCustomID.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{custom_id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert response_schema['data']['properties']['id']['type'] == 'string'
    assert response_schema['data']['properties']['id']['format'] == 'int32'
    assert 'type' in response_schema['data']['properties']
    assert 'attributes' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name', 'archived']
    assert 'relationships' in response_schema['data']['properties']
    relationships_schema = response_schema['data']['properties']['relationships']['properties']
    assert list(relationships_schema.keys()) == ['members', 'owner-member']
    members_schema = relationships_schema['members']['properties']
    assert members_schema['data']['items']['properties']['id']['type'] == 'string'
    assert members_schema['data']['items']['properties']['id']['format'] == 'int64'
    assert members_schema['data']['items']['properties']['type']['pattern'] == 'member-with-custom-ids'
    owner_member_schema = relationships_schema['owner-member']['properties']
    assert owner_member_schema['data']['properties']['id']['type'] == 'string'
    assert owner_member_schema['data']['properties']['id']['format'] == 'int64'
    assert owner_member_schema['data']['properties']['type']['pattern'] == 'member-with-custom-ids'


@pytest.mark.parametrize(
    'read_only', (
        True,
        False,
    )
)
def test_related_resource__reverse(read_only):
    """
    Correctly select id from non default pk field for both model and related models
    """
    class MemberSerializer(serializers.ModelSerializer):
        class Meta:
            model = MemberWithCustomID
            fields = ('custom_id', 'first_name', 'last_name', 'projects', 'owned_projects')
            read_only_fields = ['projects', 'owned_projects'] if read_only else []

    class MemberViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = MemberWithCustomID.objects.all()
        serializer_class = MemberSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'members', MemberViewSet, **compatibility._basename_or_base_name('members'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/members/{custom_id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert response_schema['data']['properties']['id']['type'] == 'string'
    assert response_schema['data']['properties']['id']['format'] == 'int64'
    assert 'type' in response_schema['data']['properties']
    assert 'attributes' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['first-name', 'last-name']
    assert 'relationships' in response_schema['data']['properties']
    relationships_schema = response_schema['data']['properties']['relationships']['properties']
    assert list(relationships_schema.keys()) == ['projects', 'owned-projects']
    assert relationships_schema['projects']['properties']['data']['items']['properties']['id']['type'] == 'string'
    assert relationships_schema['projects']['properties']['data']['items']['properties']['id']['format'] == 'int32'
    assert relationships_schema['owned-projects']['properties']['data']['items']['properties']['id']['type'] == 'string'
    assert relationships_schema['owned-projects']['properties']['data']['items']['properties']['id']['format'] == \
        'int32'


class ProjectWithCustomIDAndExtraProperties(ProjectWithCustomID):

    class Meta:
        app_label = 'tests'

    @property
    def one_member(self):
        return

    @property
    def many_members(self):
        return


@pytest.mark.parametrize(
    'serializer_field,expect_array', (
        (relations.SerializerMethodResourceRelatedField(model=MemberWithCustomID, source='get_member', read_only=True),
         False),
        (relations.SerializerMethodResourceRelatedField(model=MemberWithCustomID, source='get_members', read_only=True,
                                                        many=True),
         True),
        (relations.ResourceRelatedField(model=MemberWithCustomID, source='one_member', read_only=True),
         False),
        (relations.ResourceRelatedField(model=MemberWithCustomID, source='many_members', read_only=True, many=True),
         True),
        (relations.ResourceRelatedField(queryset=MemberWithCustomID.objects.all(), source='one_member'),
         False),
        (relations.ResourceRelatedField(queryset=MemberWithCustomID.objects.all(), source='many_members', many=True),
         True),
    )
)
def test_force_related_resource(serializer_field, expect_array):
    """
    Support off combinations of related resources fields – they do supply models or don't.
    """
    class ProjectSerializer(serializers.ModelSerializer):
        member_relation = serializer_field

        class Meta:
            model = ProjectWithCustomIDAndExtraProperties
            fields = ['name', 'archived', 'member_relation']

        def get_member(self):
            pass

        def get_members(self):
            pass

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = ProjectWithCustomID.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{custom_id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert response_schema['data']['properties']['id']['type'] == "string"
    assert response_schema['data']['properties']['id']['format'] == "int32"
    assert 'type' in response_schema['data']['properties']
    assert 'attributes' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name', 'archived']
    assert 'relationships' in response_schema['data']['properties']
    relation_schema = response_schema['data']['properties']['relationships']['properties']['member-relation']

    if expect_array:
        assert 'items' in relation_schema['properties']['data']
        data_schema = relation_schema['properties']['data']['items']['properties']
    else:
        assert 'properties' in relation_schema['properties']['data']
        data_schema = relation_schema['properties']['data']['properties']

    assert data_schema['id']['type'] == 'string'
    assert data_schema['id']['format'] == 'int64'


def test_id_based_on_pk():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = ProjectWithCustomID
            fields = ['name']

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = ProjectWithCustomID.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{custom_id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert response_schema['data']['properties']['id']['type'] == 'string'
    assert 'type' in response_schema['data']['properties']
    assert 'attributes' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name']



from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from rest_framework import mixins
from rest_framework import routers
from rest_framework import viewsets
from rest_framework_json_api import parsers
from rest_framework_json_api import renderers
from rest_framework_json_api import serializers

from drf_yasg_json_api.deprecation import DrfYasgJsonApiDeprecationWarning
from tests import base
from tests import compatibility
from tests import models as test_models


def test_fallback_to_rest_api():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert list(response_schema.keys()) == ['id', 'name', 'archived', 'members']


def test_deprecate_universal_inline_serializer(recwarn):
    import drf_yasg.inspectors
    import drf_yasg_json_api.inspectors

    class SwaggerAutoSchema(drf_yasg_json_api.inspectors.SwaggerAutoSchema):
        field_inspectors = [
            drf_yasg_json_api.inspectors.NamesFormatFilter,
            drf_yasg_json_api.inspectors.InlineSerializerSmartInspector,
            drf_yasg_json_api.inspectors.IntegerIDFieldInspector,
            drf_yasg_json_api.inspectors.IntegerPrimaryKeyRelatedFieldInspector,
            drf_yasg_json_api.inspectors.ManyRelatedFieldInspector,
            drf_yasg.inspectors.RelatedFieldInspector,
            drf_yasg.inspectors.SimpleFieldInspector,
            drf_yasg.inspectors.StringDefaultFieldInspector,
        ]

    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = SwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert 'type' in response_schema['data']['properties']
    assert 'attributes' in response_schema['data']['properties']

    recwarn.pop(DrfYasgJsonApiDeprecationWarning)



import logging
import warnings

from collections import OrderedDict

from django.db import models
from drf_yasg import inspectors
from drf_yasg import openapi
from drf_yasg.inspectors.field import get_model_field
from drf_yasg.inspectors.field import get_parent_serializer
from drf_yasg.utils import filter_none
from rest_framework import relations
from rest_framework import serializers
from rest_framework.serializers import BaseSerializer
from rest_framework.settings import api_settings
from rest_framework_json_api import serializers as dja_serializers
from rest_framework_json_api import utils as json_api_utils
from rest_framework_json_api.settings import json_api_settings
from rest_framework_json_api.utils import format_value
from rest_framework_json_api.utils import get_resource_name
from rest_framework_json_api.utils import get_resource_type_from_model
from rest_framework_json_api.utils import get_resource_type_from_serializer

from drf_yasg_json_api.deprecation import DrfYasgJsonApiDeprecationWarning
from drf_yasg_json_api.utils import get_field_by_source
from drf_yasg_json_api.utils import get_field_related_model
from drf_yasg_json_api.utils import get_serializer_model_primary_key
from drf_yasg_json_api.utils import is_json_api
from drf_yasg_json_api.utils import is_json_api_request
from drf_yasg_json_api.utils import is_json_api_response
from drf_yasg_json_api.utils import is_many_related_field

logger = logging.getLogger(__name__)

__all__ = [
    'InlineSerializerInspector',
    'InlineSerializerSmartInspector',
    'JSONAPISerializerInspector',
    'JSONAPISerializerSmartInspector',
    'IntegerPrimaryKeyRelatedFieldInspector',
    'IntegerIDFieldInspector',
    'ManyRelatedFieldInspector',
    'NamesFormatFilter',
    'XPropertiesFilter'
]


class JSONAPIDeclarationError(ValueError):
    pass


class JSONAPISerializerInspector(inspectors.InlineSerializerInspector):
    strip_read_fields_from_request = False
    strip_write_fields_from_response = False
    handle_json_api_only = True

    def get_schema(self, serializer):
        return self.probe_field_inspectors(serializer, openapi.Schema, self.use_definitions, is_request=False)

    def get_request_schema(self, serializer):
        return self.probe_field_inspectors(serializer, openapi.Schema, self.use_definitions, is_request=True)

    def get_included_schema(self, serializer):
        return self.probe_field_inspectors(serializer, openapi.Schema, self.use_definitions, included=True)

    def field_to_swagger_object(self, field, swagger_object_type, use_references, included=False, is_request=None,
                                **kwargs):
        if not self.is_json_api_root_serializer(field, is_request):
            if self.handle_json_api_only:
                return inspectors.NotHandled
            else:
                return super().field_to_swagger_object(field, swagger_object_type, use_references, **kwargs)

        if is_request is None and included:
            is_request = False

        if included:
            resource_name = get_resource_type_from_serializer(field)
        else:
            resource_name = get_resource_name(context={'view': self.view})

        SwaggerType, ChildSwaggerType = self._get_partial_types(field, swagger_object_type, use_references, **kwargs)
        return self.build_serializer_schema(field, resource_name, SwaggerType, ChildSwaggerType, use_references,
                                            is_request)

    def build_serializer_schema(self, serializer, resource_name, SwaggerType, ChildSwaggerType, use_references,
                                is_request=None):
        fields = json_api_utils.get_serializer_fields(serializer)
        is_post = self.method.lower() == 'post'

        id_field = self.extract_id_field(fields, serializer)
        if id_field is None and not (is_request and is_post):
            logging.warning('{view}.{serializer} does not contain id field as every resource should'.format(
                view=self.view.__class__.__name__, serializer=serializer.__class__.__name__
            ))

        attributes, req_attrs = self.extract_attributes(id_field, fields, ChildSwaggerType, use_references, is_request)
        relationships, req_rels = self.extract_relationships(fields, ChildSwaggerType, use_references, is_request)
        links = self.extract_links(fields, ChildSwaggerType, use_references) if not is_request else None

        schema_fields = filter_none(OrderedDict(
            type=self.build_type_schema(resource_name),
            id=self.probe_field_inspectors(id_field, ChildSwaggerType, use_references)
            if id_field and not (self.strip_read_fields_from_request and is_request and is_post) else None,
            attributes=openapi.Schema(type=openapi.TYPE_OBJECT, properties=attributes, required=req_attrs)
            if attributes else None,
            relationships=openapi.Schema(type=openapi.TYPE_OBJECT, properties=relationships, required=req_rels)
            if relationships else None,
            links=openapi.Schema(type=openapi.TYPE_OBJECT, properties=links)
            if links else None
        ))

        required_properties = None
        if self.is_request_or_unknown(is_request):
            required_properties = filter_none([
                'type',
                'id' if 'id' in schema_fields else None,
                'attributes' if req_attrs else None,
                'relationships' if req_rels else None,
            ])

        return SwaggerType(
            type=openapi.TYPE_OBJECT,
            properties=schema_fields,
            required=required_properties
        )

    def build_type_schema(self, resource_name, read_only=None):
        return openapi.Schema(**filter_none(OrderedDict(
            type=openapi.TYPE_STRING, pattern=resource_name, read_only=read_only or None
        )))

    def extract_id_field(self, fields, serializer: serializers.Serializer):
        # Included in fields and explicitly named "id"
        if 'id' in fields:
            serializer_id = fields['id']
            model_pk = get_serializer_model_primary_key(serializer)
            if model_pk and model_pk.name != 'id' and get_field_by_source(fields.values(), model_pk.name):
                raise JSONAPIDeclarationError('if serializer includes primary key it cannot define other field as id')
            return serializer_id

        if not isinstance(serializer, serializers.ModelSerializer):
            return None

        # Included in fields, but not as "id", find by model primary key
        model_pk = get_serializer_model_primary_key(serializer)
        serializer_id = get_field_by_source(fields.values(), model_pk.name)
        if serializer_id:
            return serializer_id

        # Not included in fields, create "temporary" field based on model primary key
        id_field_class, id_field_kwargs = serializer.build_standard_field('id', model_pk)
        serializer_id: serializers.Field = id_field_class(**id_field_kwargs,
                                                          source=model_pk.name if model_pk.name != 'id' else None)
        # NOTE: emulating binding, this is one-way binding
        # This field is safe to use and pass anywhere, but it won't be visible from serializer
        serializer_id.bind('id', serializer)
        return serializer_id

    def extract_attributes(self, id_field, fields, ChildSwaggerType, use_references, is_request=None):
        attrs = {}
        required_attrs = []
        for field_name, field in fields.items():
            if self.should_strip_from_schema(field, is_request):
                continue
            # ID is always provided in the root of JSON API so remove it from attributes
            if id_field and field_name == id_field.field_name:
                continue
            # Skip fields with relations
            if isinstance(field, (relations.RelatedField, relations.ManyRelatedField, BaseSerializer)):
                continue

            attrs[field_name] = self.probe_field_inspectors(field, ChildSwaggerType, use_references)
            if self.is_request_or_unknown(is_request) and field.required and not field.read_only:
                required_attrs.append(field_name)
        return attrs, (required_attrs or None)

    def extract_relationships(self, fields, ChildSwaggerType, use_references, is_request=None):
        relationships = OrderedDict()
        required_relationships = []
        for field_name, field in fields.items():
            self.maybe_fix_broken_parent_relation(field)

            if self.should_strip_from_schema(field, is_request):
                continue
            # Self url field
            if field_name == api_settings.URL_FIELD_NAME:
                continue
            # Skip fields without relations
            if not isinstance(field, (relations.RelatedField, relations.ManyRelatedField, serializers.Serializer)):
                continue

            # Produce swagger output
            relation_data_schema = openapi.Schema(**filter_none(OrderedDict(
                type=openapi.TYPE_OBJECT,
                properties=OrderedDict(
                    id=self.probe_field_inspectors(field, ChildSwaggerType, use_references),
                    type=self.build_type_schema(
                        self.get_resource_name_from_related_id_field(field_name, field),
                        read_only=field.read_only
                    ),
                ),
                required=['id', 'type'] if (self.is_request_or_unknown(is_request)) and not field.read_only else None,
            )))

            if is_many_related_field(field):
                relation_data_schema = openapi.Schema(type=openapi.TYPE_ARRAY, items=relation_data_schema)

            relation_links_schema = self.get_links_from_id_field(field_name, field)
            if relation_links_schema:
                relation_links_schema = openapi.Schema(type=openapi.TYPE_OBJECT, properties=relation_links_schema)

            is_relation_required = self.is_request_or_unknown(is_request) and field.required and not field.read_only

            relationships[field_name] = openapi.Schema(**filter_none(OrderedDict(
                type=openapi.TYPE_OBJECT,
                properties=filter_none({
                    'data': relation_data_schema,
                    'links': relation_links_schema if self.not_request_or_unknown(is_request) else None
                }),
                required=['data'] if is_relation_required else None,
                read_only=field.read_only or None,
                x_read_only=field.read_only or None,
            )))
            if is_relation_required:
                required_relationships.append(field_name)

        return relationships, (required_relationships or None)

    def extract_links(self, fields, ChildSwaggerType, use_references):
        self_field_name = api_settings.URL_FIELD_NAME

        return filter_none(OrderedDict(
            self=openapi.Schema(type=openapi.TYPE_STRING, format=openapi.FORMAT_URI)
            if self_field_name in fields and isinstance(fields[self_field_name], serializers.RelatedField) else None
        ))

    def get_resource_name_from_related_id_field(self, field_name, id_field):
        # Unpack ManyRelatedField from many wrapper
        id_field = getattr(id_field, 'child_relation', None) or id_field

        # Not very frequent case but different from others
        if isinstance(id_field, serializers.Serializer):
            return json_api_utils.get_resource_type_from_serializer(id_field)
        # Most cases
        else:
            # Try to get from included serializers
            parent_serializer = get_parent_serializer(id_field)
            if isinstance(id_field, dja_serializers.ResourceRelatedField):
                resource_name = id_field.get_resource_type_from_included_serializer()
                if resource_name:
                    return resource_name

            related_model = get_field_related_model(id_field)
            if related_model:
                return get_resource_type_from_model(related_model)

        raise ValueError(f"Unable to extract resource name for {parent_serializer}.{field_name} serializer field")

    def get_links_from_id_field(self, field_name, id_field):
        links = OrderedDict()
        if isinstance(id_field, dja_serializers.ResourceRelatedField):
            if id_field.related_link_lookup_field is not None:
                links['related'] = openapi.Schema(type=openapi.TYPE_STRING, pattern=openapi.FORMAT_URI, read_only=True)
            if id_field.self_link_view_name is not None:
                links['self'] = openapi.Schema(type=openapi.TYPE_STRING, pattern=openapi.FORMAT_URI, read_only=True)
        return links or None

    def is_json_api_root_serializer(self, field, is_request=False):
        return field and field.parent is None and isinstance(field, serializers.Serializer) and (
            (not is_request and is_json_api_response(self.view.renderer_classes))
            or ((is_request or is_request is None) and is_json_api_request(self.view.parser_classes))
        )

    def should_strip_from_schema(self, field, is_request):
        return (self.strip_read_fields_from_request and is_request and field.read_only) or \
               (self.strip_write_fields_from_response and is_request is False and field.write_only)

    def not_request_or_unknown(self, is_request):
        # evaluate None to True as well
        return not is_request

    def is_request_or_unknown(self, is_request):
        # evaluate None to True as well
        return is_request is None or is_request

    def maybe_fix_broken_parent_relation(self, candidate_field):
        """
        SerializerMethodResourceRelatedField is a bit hacky and breaks field.parent.parent...serializer chain when used
        with many=True. To avoid multiple exception in random inspectors, we just check every relation if fix is needed.
        """
        child_field = getattr(candidate_field, 'child_relation', None)
        if child_field and not getattr(child_field, 'parent', None):
            setattr(child_field, 'parent', candidate_field)


class JSONAPISerializerSmartInspector(JSONAPISerializerInspector):
    strip_read_fields_from_request = True
    strip_write_fields_from_response = True


class InlineSerializerInspector(JSONAPISerializerInspector):
    handle_json_api_only = False

    def __init__(self, *args, **kwargs):
        warnings.warn(
            f'{self.__class__.__name__} is deprecated, use drf_yasg_json_api.inspectors.JSONAPISerializerInspector '
            'instead to render docs for JSON API views and optionally '
            'drf_yasg.inspectors.ReferencingSerializerInspector/InlineSerializerInspector to render docs for '
            'non JSON API views',
            DrfYasgJsonApiDeprecationWarning
        )
        super().__init__(*args, **kwargs)


class InlineSerializerSmartInspector(JSONAPISerializerSmartInspector):
    handle_json_api_only = False

    def __init__(self, *args, **kwargs):
        warnings.warn(
            f'{self.__class__.__name__} is deprecated, use drf_yasg_json_api.inspectors.JSONAPISerializerSmartInspector'
            ' instead to render docs for JSON API views and optionally '
            'drf_yasg.inspectors.ReferencingSerializerInspector/InlineSerializerInspector to render docs for '
            'non JSON API views',
            DrfYasgJsonApiDeprecationWarning
        )
        super().__init__(*args, **kwargs)


class IntegerFieldInspectorMixin:
    int64_fields = (models.BigIntegerField, models.BigAutoField)

    def get_format(self, model_field):
        return openapi.FORMAT_INT64 if isinstance(model_field, self.int64_fields) else openapi.FORMAT_INT32


class IntegerIDFieldInspector(IntegerFieldInspectorMixin, inspectors.FieldInspector):
    """
    Force string type on ID field that on model level is integer. Since we get here just an integer, we look for:
     - Primary Key of model that is models.IntegerField
     - Serializer field named "id" that is serializers.IntegerField
    """

    def field_to_swagger_object(self, field, swagger_object_type, **kwargs):
        if not isinstance(field, serializers.IntegerField) or not is_json_api(self.view):
            return inspectors.NotHandled

        stringify_id = False
        integer_format = None

        parent_serializer = get_parent_serializer(field)
        serializer_meta = getattr(parent_serializer, 'Meta', None)
        model = getattr(serializer_meta, 'model', None)
        if model is not None:
            field_name = getattr(field, 'source', None) or field.field_name
            model_field = get_model_field(model, field_name)
            # Check for primary key only if sure it is pure Field (not FieldCacheMixin or anything)
            if (
                model_field is not None and
                isinstance(model_field, (models.IntegerField, models.AutoField)) and
                model_field.primary_key
            ):
                stringify_id = True
                integer_format = self.get_format(model_field)

        elif field.field_name == 'id' and isinstance(field, serializers.IntegerField):
            stringify_id = True

        if not stringify_id:
            return inspectors.NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(field, swagger_object_type, **kwargs)
        return SwaggerType(type=openapi.TYPE_STRING, format=integer_format)


class IntegerPrimaryKeyRelatedFieldInspector(IntegerFieldInspectorMixin, inspectors.FieldInspector):
    """
    Force string type on PrimaryRelatedField that refers to model with integer primary key.
    """

    def field_to_swagger_object(self, field, swagger_object_type, **kwargs):
        if not isinstance(field, serializers.PrimaryKeyRelatedField) or not is_json_api(self.view):
            return inspectors.NotHandled

        if is_many_related_field(field):
            return inspectors.NotHandled

        related_model = get_field_related_model(field)
        if related_model:
            related_model_pk_field = get_model_field(related_model, 'pk')
            if isinstance(related_model_pk_field, (models.IntegerField, models.AutoField)):
                SwaggerType, ChildSwaggerType = self._get_partial_types(field, swagger_object_type, **kwargs)
                return SwaggerType(type=openapi.TYPE_STRING, format=self.get_format(related_model_pk_field))

        return inspectors.NotHandled


class ManyRelatedFieldInspector(inspectors.SimpleFieldInspector):
    """
    Unwrap ManyRelatedField child relation as array node has already been added by SerializerInspector.
    """

    def field_to_swagger_object(self, field, swagger_object_type, **kwargs):
        if is_many_related_field(field) and is_json_api(self.view):
            return self.probe_field_inspectors(field.child_relation, swagger_object_type, **kwargs)

        return inspectors.NotHandled


class NamesFormatFilter(inspectors.FieldInspector):

    def format_string(self, s):
        return format_value(s, json_api_settings.FORMAT_FIELD_NAMES)

    def format_schema(self, schema):
        """Recursively format property names for the given schema according to``JSON_API_FORMAT_KEYS`` setting.
        The target schema object must be modified in-place.

        :param openapi.Schema schema: the :class:`.Schema` object
        """
        if getattr(schema, 'properties', {}):
            schema.properties = OrderedDict(
                (self.format_string(key), self.format_schema(openapi.resolve_ref(val, self.components)) or val)
                for key, val in schema.properties.items()
            )

            if getattr(schema, 'required', []):
                schema.required = [self.format_string(p) for p in schema.required]

    def process_result(self, result, method_name, obj, **kwargs):
        if isinstance(result, openapi.Schema.OR_REF) and is_json_api(self.view):
            schema = openapi.resolve_ref(result, self.components)
            self.format_schema(schema)

        return result


class XPropertiesFilter(inspectors.FieldInspector):

    def add_write_only(self, result, obj):
        if obj.write_only:
            result.x_write_only = True

    def fix_read_only(self, result, obj):
        # drf_yasg is very cautious about setting read_only, only leaf nodes are allowed to be read only,
        # but in some cases it goes too far and some leaf nodes cases are omitted too.
        # This workaround fixes obvious cases – types that are always leaf nodes
        if obj.read_only:
            if isinstance(result, openapi.Schema) \
                    and result.type in (openapi.TYPE_STRING, openapi.TYPE_INTEGER, openapi.TYPE_BOOLEAN):
                result.read_only = True
            # For non-leaf nodes make read only visible by adding x prefix to avoid conflict with OpenApi 2 validation
            else:
                result.x_read_only = True

    def process_result(self, result, method_name, obj, **kwargs):
        if result is not None:
            self.add_write_only(result, obj)
            self.fix_read_only(result, obj)
        return result



from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from rest_framework import mixins
from rest_framework import routers
from rest_framework import viewsets
from rest_framework_json_api import parsers
from rest_framework_json_api import renderers
from rest_framework_json_api import serializers

from tests import base
from tests import compatibility
from tests import models as test_models


class IncludedStringPathMemberSerializer(serializers.ModelSerializer):
    # projects = serializers.ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = test_models.Member
        fields = ['first_name', 'last_name', 'projects']


def test_included__string_path():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

        included_serializers = {
            'members': 'tests.test_schema.test_get_included.IncludedStringPathMemberSerializer',
        }

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'included' in response_schema
    assert 'members' in response_schema['included']['properties']

    request_parameters_schema = swagger['paths']['/projects/{id}/']['get']['parameters']
    assert request_parameters_schema[0]['name'] == 'include'
    assert request_parameters_schema[0]['description'].endswith(': members')


class IncludedRecursiveMemberSerializer(serializers.ModelSerializer):
    # projects = serializers.ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = test_models.Member
        fields = ['first_name', 'last_name', 'projects']

    included_serializers = {
        'projects': 'tests.test_schema.test_get_included.IncludedRecursiveProjectSerializer',
    }


class IncludedRecursiveProjectSerializer(serializers.ModelSerializer):
    class Meta:
        model = test_models.Project
        fields = ('id', 'name', 'archived', 'members', 'sub_projects')

    included_serializers = {
        'members': 'tests.test_schema.test_get_included.IncludedRecursiveMemberSerializer',
        'sub_projects': 'self',
    }


def test_included__recursive():
    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = IncludedRecursiveProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'included' in response_schema
    assert 'members' in response_schema['included']['properties']
    assert 'projects' in response_schema['included']['properties']
    included_members_schema = response_schema['included']['properties']['members']['properties']
    assert 'projects' in included_members_schema['relationships']['properties']
    included_projects_schema = response_schema['included']['properties']['projects']['properties']
    assert 'sub-projects' in included_projects_schema['relationships']['properties']
    assert 'members' in included_projects_schema['relationships']['properties']

    request_parameters_schema = swagger['paths']['/projects/{id}/']['get']['parameters']
    assert request_parameters_schema[0]['name'] == 'include'
    assert request_parameters_schema[0]['description'].endswith(
        ': sub-projects [recursive], members, members.projects [recursive through: members.projects]'
    )



import operator

from unittest import mock

import drf_yasg.inspectors

from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from drf_yasg.utils import swagger_auto_schema
from rest_framework import mixins
from rest_framework import routers
from rest_framework import views
from rest_framework import viewsets
from rest_framework_json_api import django_filters
from rest_framework_json_api import filters
from rest_framework_json_api import pagination
from rest_framework_json_api import parsers
from rest_framework_json_api import renderers
from rest_framework_json_api import serializers

import drf_yasg_json_api.inspectors

from tests import base
from tests import compatibility
from tests import models as test_models

try:
    from django.urls import path
except ImportError:
    # Django 1.11 compatible
    from django.conf.urls import url as path


def test_get():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert response_schema['data']['properties']['id']['type'] == 'string'
    assert 'type' in response_schema['data']['properties']
    assert response_schema['data']['properties']['type']['pattern'] == 'projects'
    assert 'attributes' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name', 'archived']
    assert 'relationships' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['relationships']['properties'].keys()) == ['members']
    members_schema = response_schema['data']['properties']['relationships']['properties']['members']['properties']
    assert members_schema['data']['items']['properties']['id']['type'] == 'string'
    assert members_schema['data']['items']['properties']['type']['pattern'] == 'members'


def test_pagination():
    class SwaggerAutoSchemaWithPagination(base.BasicSwaggerAutoSchema):
        paginator_inspectors = [
            drf_yasg_json_api.inspectors.DjangoRestResponsePagination,
            drf_yasg.inspectors.DjangoRestResponsePagination,
            drf_yasg.inspectors.CoreAPICompatInspector,
        ]

    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = SwaggerAutoSchemaWithPagination
        pagination_class = pagination.JsonApiPageNumberPagination

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_parameters_schema = swagger['paths']['/projects/']['get']['parameters']
    assert set(map(operator.itemgetter('name'), request_parameters_schema)) == {'page[number]', 'page[size]'}

    response_schema = swagger['paths']['/projects/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['items']['properties']
    assert 'type' in response_schema['data']['items']['properties']
    assert 'attributes' in response_schema['data']['items']['properties']
    assert 'relationships' in response_schema['data']['items']['properties']
    assert 'links' in response_schema
    assert set(response_schema['links']['properties'].keys()) == {'first', 'next', 'last', 'prev'}
    assert 'meta' in response_schema
    assert 'pagination' in response_schema['meta']['properties']
    pagination_response_schema = response_schema['meta']['properties']['pagination']['properties']
    assert set(pagination_response_schema.keys()) == {'page', 'pages', 'count'}


@mock.patch('drf_yasg_json_api.inspectors.view.logger')
def test_list_missing_serializer_warning(logger):
    class ProjectView(views.APIView):
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

        def get(*args, **kwargs):
            pass

    urlpatterns = [
        path('projects/', ProjectView.as_view())
    ]

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=urlpatterns)

    swagger = generator.get_schema(request=None, public=True)

    assert swagger['paths']['/projects/']['get']
    assert len(logger.warning.mock_calls) == 2
    logger.warning.assert_called_with('Missing schema definition for list action of ProjectView, '
                                      'have you defined get_serializer?')


def test_strip_write_only():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')
            extra_kwargs = {
                'name': {'write_only': False},
                'archived': {'write_only': True},
                'members': {'write_only': True},
            }

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'id' in response_schema['data']['properties']
    assert 'type' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name']
    assert 'relationships' not in response_schema['data']['properties']


@mock.patch('rest_framework.settings.api_settings.URL_FIELD_NAME', 'obj_url')
def test_data_links_self():

    class ProjectSerializer(serializers.ModelSerializer):
        obj_url = serializers.HyperlinkedIdentityField(view_name='any')

        class Meta:
            model = test_models.Project
            fields = ('id', 'obj_url')

    class ProjectViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']

    assert 'id' in response_schema['data']['properties']
    assert 'type' in response_schema['data']['properties']
    assert 'links' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['links']['properties'].keys()) == ['self']


def test_filter():
    class FilterSwaggerAutoSchema(base.BasicSwaggerAutoSchema):
        filter_inspectors = [drf_yasg_json_api.inspectors.DjangoFilterInspector]

    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = FilterSwaggerAutoSchema

        filter_backends = (filters.QueryParameterValidationFilter, django_filters.DjangoFilterBackend)
        filterset_fields = {
            'archived': ('exact',),
        }

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_parameters_schema = swagger['paths']['/projects/']['get']['parameters']
    assert request_parameters_schema[0]['name'] == 'filter[archived]'


def test_non_model():
    class ProjectSerializer(serializers.Serializer):
        id = serializers.IntegerField()
        name = serializers.CharField()

    class ProjectViewSet(viewsets.ViewSet):
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema
        resource_name = 'projects'
        serializer_class = ProjectSerializer

        def get_serializer(self, *args, **kwargs):
            return self.serializer_class(*args, **kwargs)

        def retrieve(self, request):
            pass

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'properties' in response_schema['data']
    assert 'id' in response_schema['data']['properties']
    assert 'type' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name']


def test_non_model_many():
    class ProjectSerializer(serializers.Serializer):
        id = serializers.IntegerField()
        name = serializers.CharField()

    class ProjectViewSet(viewsets.ViewSet):
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema
        resource_name = 'projects'
        serializer_class = ProjectSerializer

        def get_serializer(self, *args, **kwargs):
            return self.serializer_class(*args, **kwargs)

        def list(self, request):
            pass

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/']['get']['responses']['200']['schema']['properties']
    assert 'items' in response_schema['data']
    assert 'id' in response_schema['data']['items']['properties']
    assert 'type' in response_schema['data']['items']['properties']
    assert list(response_schema['data']['items']['properties']['attributes']['properties'].keys()) == ['name']


def test_non_model__responses_override():
    class ProjectSerializer(serializers.Serializer):
        id = serializers.IntegerField()
        name = serializers.CharField()

    class ProjectViewSet(viewsets.ViewSet):
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema
        resource_name = 'projects'

        @swagger_auto_schema(responses={200: ProjectSerializer()})
        def retrieve(self, request):
            pass

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/{id}/']['get']['responses']['200']['schema']['properties']
    assert 'properties' in response_schema['data']
    assert 'id' in response_schema['data']['properties']
    assert 'type' in response_schema['data']['properties']
    assert list(response_schema['data']['properties']['attributes']['properties'].keys()) == ['name']


def test_non_model_many__responses_override():
    class ProjectSerializer(serializers.Serializer):
        id = serializers.IntegerField()
        name = serializers.CharField()

    class ProjectViewSet(viewsets.ViewSet):
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema
        resource_name = 'projects'

        @swagger_auto_schema(responses={200: ProjectSerializer(many=True)})
        def list(self, request):
            pass

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    response_schema = swagger['paths']['/projects/']['get']['responses']['200']['schema']['properties']
    assert 'items' in response_schema['data']
    assert 'id' in response_schema['data']['items']['properties']
    assert 'type' in response_schema['data']['items']['properties']
    assert list(response_schema['data']['items']['properties']['attributes']['properties'].keys()) == ['name']



import copy
import itertools

from typing import Optional

from django.db import models
from django.utils.module_loading import import_string
from drf_yasg.inspectors.field import get_parent_serializer
from rest_framework import serializers


def is_json_api(view):
    return is_json_api_response(view.renderer_classes) or is_json_api_request(view.parser_classes)


def is_json_api_response(renderer_classes):
    from rest_framework_json_api.renderers import JSONRenderer as JSONAPIRenderer
    return any(issubclass(renderer, JSONAPIRenderer) for renderer in renderer_classes)


def is_json_api_request(parser_classes):
    from rest_framework_json_api.parsers import JSONParser as JSONAPIParser
    return any(issubclass(parser, JSONAPIParser) for parser in parser_classes)


def get_related_model(model, source):
    if source == '*':
        return model

    descriptor = model
    try:
        for attr in source.split('.'):
            descriptor = getattr(descriptor, attr)
    except AttributeError:  # pragma: no cover
        return None

    try:
        is_forward = descriptor.field in itertools.chain(model._meta.fields, model._meta.many_to_many)
    except AttributeError:  # pragma: no cover
        return None

    if is_forward:
        return descriptor.field.related_model
    else:
        return descriptor.field.model


def get_serializer_model_primary_key(serializer):
    if not isinstance(serializer, serializers.ModelSerializer):
        return None  # pragma: no cover
    return [f for f in serializer.Meta.model._meta.fields if f.primary_key][0]


def get_field_by_source(fields: list, source):
    for field in fields:
        if field.source == source:
            return field
    return None


def get_field_related_model(field) -> Optional[models.Model]:
    # Try extracting directly from field
    related_model = _get_field_model(field)
    # If failed try to extract by traversing model and model fields
    if related_model is None:
        parent_serializer = get_parent_serializer(field)
        serializer_meta = getattr(parent_serializer, 'Meta', None)
        model = getattr(serializer_meta, 'model', None)
        if model is not None:
            related_model = get_related_model(model, source=get_field_source(field))
    return related_model


def _get_field_model(field: serializers.Field) -> Optional[models.Model]:
    field_model = getattr(field, 'model', None)
    if field_model:
        return field_model

    try:
        return field.queryset.model
    except AttributeError:  # pragma: no cover
        return None


def is_many_related_field(field):
    # The check for child relation attribute covers ManyRelationField as well as other possible cases like
    # hacky SerializerMethodResourceRelatedField
    return getattr(field, 'child_relation', None)


def get_field_source(field: serializers.Field):
    source = field.source or field.field_name
    # If no source and parent is not serializer it is child_field of other field
    if not source and not isinstance(field.parent, serializers.BaseSerializer):
        return field.parent.source or field.parent.field_name
    return source


def get_included_serializers(serializer):
    # Extracted from djangorestframework-jsonapi for backwards compatibility

    included_serializers = copy.copy(
        getattr(serializer, "included_serializers", dict())
    )

    for name, value in iter(included_serializers.items()):
        if not isinstance(value, type):
            if value == "self":
                included_serializers[name] = (
                    serializer if isinstance(serializer, type) else serializer.__class__
                )
            else:
                included_serializers[name] = import_string(value)

    return included_serializers



import drf_yasg.inspectors

import drf_yasg_json_api.inspectors


class BasicSwaggerAutoSchema(drf_yasg_json_api.inspectors.SwaggerAutoSchema):
    field_inspectors = [
        drf_yasg_json_api.inspectors.NamesFormatFilter,
        drf_yasg_json_api.inspectors.JSONAPISerializerSmartInspector,
        drf_yasg.inspectors.ReferencingSerializerInspector,
        drf_yasg_json_api.inspectors.IntegerIDFieldInspector,
        drf_yasg_json_api.inspectors.IntegerPrimaryKeyRelatedFieldInspector,
        drf_yasg_json_api.inspectors.ManyRelatedFieldInspector,
        drf_yasg.inspectors.RelatedFieldInspector,
        drf_yasg.inspectors.SimpleFieldInspector,
        drf_yasg.inspectors.StringDefaultFieldInspector,
    ]



import logging

from collections import OrderedDict

from django.utils.functional import cached_property
from drf_yasg import inspectors
from drf_yasg import openapi
from drf_yasg.utils import filter_none
from rest_framework_json_api import pagination

logger = logging.getLogger(__name__)

__all__ = [
    'DjangoFilterInspector',
    'DjangoRestResponsePagination',
]


class DjangoFilterInspector(inspectors.CoreAPICompatInspector):
    @cached_property
    def django_filters(self):
        from rest_framework_json_api import django_filters
        return django_filters

    def get_filter_parameters(self, filter_backend):
        if not isinstance(filter_backend, self.django_filters.DjangoFilterBackend):
            return inspectors.NotHandled
        return super().get_filter_parameters(filter_backend)

    def coreapi_field_to_parameter(self, field):
        parameter = super().coreapi_field_to_parameter(field)
        parameter.name = f'filter[{parameter.name}]'
        return parameter


class DjangoRestResponsePagination(inspectors.PaginatorInspector):

    def get_paginated_response(self, paginator, response_schema):
        assert 'data' in response_schema['properties'], "expected data field in response"
        assert response_schema['properties']['data'].type == openapi.TYPE_ARRAY, "array expected for paged response"
        if not isinstance(paginator, (pagination.JsonApiPageNumberPagination, pagination.JsonApiLimitOffsetPagination)):
            return inspectors.NotHandled

        has_page = isinstance(paginator, pagination.JsonApiPageNumberPagination)
        meta_schema = openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties=OrderedDict(
                pagination=openapi.Schema(
                    type=openapi.TYPE_OBJECT,
                    properties=OrderedDict(filter_none((
                        ('count', openapi.Schema(type=openapi.TYPE_INTEGER)),
                        ('page', openapi.Schema(type=openapi.TYPE_INTEGER) if has_page else None),
                        ('pages', openapi.Schema(type=openapi.TYPE_INTEGER) if has_page else None),
                        ('limit', openapi.Schema(type=openapi.TYPE_INTEGER) if not has_page else None),
                        ('offset', openapi.Schema(type=openapi.TYPE_INTEGER) if not has_page else None),
                    ))),
                )
            )
        )
        links_schema = openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties=OrderedDict((
                ('first', openapi.Schema(type=openapi.TYPE_STRING, format=openapi.FORMAT_URI)),
                ('next', openapi.Schema(type=openapi.TYPE_STRING, format=openapi.FORMAT_URI)),
                ('last', openapi.Schema(type=openapi.TYPE_STRING, format=openapi.FORMAT_URI)),
                ('prev', openapi.Schema(type=openapi.TYPE_STRING, format=openapi.FORMAT_URI)),
            )),
        )

        return openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties=filter_none(OrderedDict(
                links=links_schema,
                meta=meta_schema,
                data=response_schema.properties['data'],
                included=response_schema.properties.get('included')
            ))
        )



import logging

from collections import OrderedDict

from drf_yasg import inspectors
from drf_yasg import openapi
from drf_yasg.utils import filter_none
from drf_yasg.utils import guess_response_status
from rest_framework import serializers
from rest_framework.status import is_success
from rest_framework_json_api.settings import json_api_settings
from rest_framework_json_api.utils import format_value
from rest_framework_json_api.utils import get_resource_type_from_serializer

from drf_yasg_json_api.utils import get_included_serializers
from drf_yasg_json_api.utils import is_json_api_request
from drf_yasg_json_api.utils import is_json_api_response

__all__ = [
    'SwaggerAutoSchema',
]

logger = logging.getLogger(__name__)


class SwaggerAutoSchema(inspectors.SwaggerAutoSchema):
    def get_request_body_schema(self, serializer):
        """
        Hook in to generate request schema from view's serializer OR overridden using `request_body` argument of
        of `swagger_auto_schema` decorator.
        """
        schema = self.serializer_to_request_schema(serializer)
        if is_json_api_request(self.get_parser_classes()):
            if schema is not None:
                schema = openapi.Schema(
                    type=openapi.TYPE_OBJECT,
                    properties=OrderedDict({
                        'data': schema,
                    })
                )
        return schema

    def get_response_schemas(self, response_serializers):
        """
        Hook in to generate response schemas for all pure (not converted to schema) serializers which can be provided
        using `responses` argument of `swagger_auto_schema` decorator.
        """
        if not is_json_api_response(self.get_renderer_classes()):
            return super().get_response_schemas(response_serializers)

        response_schemas = OrderedDict()
        for status_code, serializer in response_serializers.items():
            if is_success(int(status_code)) and isinstance(serializer, serializers.BaseSerializer):
                response_schemas[status_code] = self.get_overridden_response_schema(serializer)
            else:
                response_schemas[status_code] = serializer

        return super().get_response_schemas(response_schemas)

    def get_overridden_response_schema(self, serializer):
        if getattr(serializer, 'many', False):
            # Use list's item as main serializer, create new instance to make it look like root serializer not child
            serializer = serializer.child.__class__()
            serializer_schema = openapi.Schema(
                type=openapi.TYPE_ARRAY,
                items=self.serializer_to_schema(serializer)
            )
        else:
            serializer_schema = self.serializer_to_schema(serializer)

        return openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties=OrderedDict(
                data=serializer_schema,
                included=self.get_included_schema_for_response(serializer)
            )
        )

    def get_default_responses(self):
        """
        Hook in to generate default response schema from view's serializer. Used only when no overriding response is
        provided using `responses` argument of `swagger_auto_schema` decorator.
        """
        if not is_json_api_response(self.get_renderer_classes()):
            return super().get_default_responses()

        method = self.method.lower()

        default_status = guess_response_status(method)
        default_schema = ''
        if method in ('get', 'post', 'put', 'patch'):
            default_serializer = self.get_default_response_serializer()

            default_schema = openapi.Schema(
                type=openapi.TYPE_OBJECT,
                properties=OrderedDict(
                    data=self.get_data_schema_for_default_response(default_serializer),
                    included=self.get_included_schema_for_response(default_serializer),
                )
            )
            if self.should_page():
                default_schema = self.get_paginated_response(default_schema)

        return filter_none(OrderedDict({str(default_status): default_schema}))

    def get_data_schema_for_default_response(self, default_serializer):
        if isinstance(default_serializer, openapi.Schema):
            default_data_schema = default_serializer
        elif default_serializer:
            default_data_schema = self.serializer_to_schema(default_serializer) or ''
        else:
            default_data_schema = ''

        if self.is_list_view() and self.method.lower() == 'get':
            if default_data_schema:
                default_data_schema = openapi.Schema(type=openapi.TYPE_ARRAY, items=default_data_schema)
            else:
                logger.warning(
                    'Missing schema definition for list action of {view_name}, have you defined get_serializer?'.format(
                        view_name=self.view.__class__.__name__
                    )
                )

        return default_data_schema

    def get_included_schema_for_response(self, serializer):
        included_paths, included_serializers = self._get_included_paths_and_serializers(serializer)
        if not included_serializers:
            return None

        return openapi.Schema(
            type=openapi.TYPE_OBJECT,
            description='note: expect this field to be an array consisting of items of types listed below',
            properties={get_resource_type_from_serializer(serializer): self.serializer_to_included_schema(serializer())
                        for i, serializer in enumerate(included_serializers)}
        )

    def serializer_to_included_schema(self, serializer):
        return self.probe_inspectors(
            self.field_inspectors, 'get_included_schema', serializer, {'field_inspectors': self.field_inspectors}
        )

    def serializer_to_request_schema(self, serializer):
        return self.probe_inspectors(
            self.field_inspectors, 'get_request_schema', serializer, {'field_inspectors': self.field_inspectors},
        )

    def get_query_parameters(self):
        """
        Hook in to add `include` parameter supported by response serializer.

        Supports generating parameter for success response from `responses` using `swagger_auto_schema` decorator
        and (of course) for view's main serializer if it's not overridden by decorator.

        """
        if not is_json_api_response(self.get_renderer_classes()):
            return super().get_query_parameters()

        success_response_serializers = [
            serializer for status_code, serializer in self.get_response_serializers().items()
            if (
                is_success(int(status_code)) and
                isinstance(serializer, serializers.BaseSerializer) and
                hasattr(serializer, 'included_serializers')
            )
        ]

        if not success_response_serializers:
            response_serializer = self.get_default_response_serializer()
        else:
            response_serializer = success_response_serializers[0]
            if len(success_response_serializers) > 1:  # pragma: no cover
                logger.warning(
                    'More than one response serializer for view {view_name} method {method} '
                    'provides included serializers, falling back to first one'.format(
                        view_name=self.view.__class__.__name__, method=self.method
                    )
                )

        return super().get_query_parameters() + self.get_query_parameters_included(response_serializer)

    def get_query_parameters_included(self, serializer):
        parameters = []

        if hasattr(serializer, 'included_serializers'):
            paths, serializers = self._get_included_paths_and_serializers(serializer)
            parameters.append(openapi.Parameter(
                type=openapi.TYPE_STRING,
                in_=openapi.IN_QUERY,
                name='include',
                description='Include relations in response. Available relations: {relation_paths}'.format(
                    relation_paths=", ".join(paths)
                ),
                format='comma-separated-array'
            ))

        return parameters

    MAX_INCLUDED_PATH_DEPTH = 20

    def _get_included_paths_and_serializers(self, field):
        field_cls = field if isinstance(field, type) else field.__class__
        all_included_paths = []
        all_included_serializers = set()
        serializers_to_visit = [([], [], field_cls)]
        while serializers_to_visit:
            path, parent_serializers, serializer = serializers_to_visit.pop()
            if len(path) > self.MAX_INCLUDED_PATH_DEPTH:  # pragma: no cover
                logger.warning('Exceeded max included path limit ({limit}), ignoring longer paths'.format(
                    limit=self.MAX_INCLUDED_PATH_DEPTH
                ))
                continue

            # Support recursive reference using "self" keyword or indirect recursion using lazy string paths
            if serializer in parent_serializers:
                if parent_serializers[-1] is serializer:
                    all_included_paths.append('{path} [recursive]'.format(path=".".join(path)))
                else:
                    recursive_path = path[parent_serializers.index(serializer):]
                    all_included_paths.append('{path} [recursive through: {recursive_path}]'.format(
                        path=".".join(path), recursive_path=".".join(recursive_path))
                    )
                continue
            if path:
                all_included_paths.append(".".join(path))
            included_serializers = get_included_serializers(serializer)
            for name, sub_serializer in included_serializers.items():
                all_included_serializers.add(sub_serializer)
                serializers_to_visit.append((
                    path + [self._format_key(name)],
                    parent_serializers + [serializer],
                    sub_serializer
                ))

        return all_included_paths, all_included_serializers

    def _format_key(self, s):
        return format_value(s, json_api_settings.FORMAT_FIELD_NAMES)



import drf_yasg.inspectors

from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from rest_framework import mixins
from rest_framework import routers
from rest_framework import viewsets
from rest_framework_json_api import parsers
from rest_framework_json_api import renderers
from rest_framework_json_api import serializers

import drf_yasg_json_api.inspectors

from tests import base
from tests import compatibility
from tests import models as test_models


def test_post():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_body_schema = swagger['paths']['/projects/']['post']['parameters'][0]['schema']['properties']
    assert 'id' not in request_body_schema['data']['properties']
    assert 'type' in request_body_schema['data']['properties']
    assert 'attributes' in request_body_schema['data']['properties']
    assert list(request_body_schema['data']['properties']['attributes']['properties'].keys()) == ['name', 'archived']
    assert 'relationships' in request_body_schema['data']['properties']
    assert list(request_body_schema['data']['properties']['relationships']['properties'].keys()) == ['members']


def test_put():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')

    class ProjectViewSet(mixins.UpdateModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_body_schema = swagger['paths']['/projects/{id}/']['put']['parameters'][0]['schema']['properties']
    assert 'id' in request_body_schema['data']['properties']
    assert 'type' in request_body_schema['data']['properties']
    assert 'attributes' in request_body_schema['data']['properties']
    assert list(request_body_schema['data']['properties']['attributes']['properties'].keys()) == ['name', 'archived']
    assert 'relationships' in request_body_schema['data']['properties']
    assert list(request_body_schema['data']['properties']['relationships']['properties'].keys()) == ['members']


def test_post__strip_read_only_fields():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')
            read_only_fields = ['archived', 'members']

    class ProjectViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_body_schema = swagger['paths']['/projects/']['post']['parameters'][0]['schema']['properties']
    assert 'id' not in request_body_schema['data']['properties']
    assert 'type' in request_body_schema['data']['properties']
    assert 'attributes' in request_body_schema['data']['properties']
    assert list(request_body_schema['data']['properties']['attributes']['properties'].keys()) == ['name']
    assert 'relationships' not in request_body_schema['data']['properties']


def test_post__mark_as_required():
    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')
            extra_kwargs = {
                'name': {'required': True},
                'archived': {'required': False},
                'members': {'required': True},
            }

    class ProjectViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = base.BasicSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_body_schema = swagger['paths']['/projects/']['post']['parameters'][0]['schema']['properties']
    assert set(request_body_schema['data']['required']) == {'type', 'attributes', 'relationships'}
    assert request_body_schema['data']['properties']['attributes']['required'] == ['name']
    assert request_body_schema['data']['properties']['relationships']['required'] == ['members']
    members_schema = request_body_schema['data']['properties']['relationships']['properties']['members']
    assert members_schema['required'] == ['data']
    assert set(members_schema['properties']['data']['items']['required']) == {'type', 'id'}


def test_post__x_properties():
    class XPropertiesSwaggerAutoSchema(drf_yasg_json_api.inspectors.SwaggerAutoSchema):
        field_inspectors = [
            drf_yasg_json_api.inspectors.NamesFormatFilter,
            drf_yasg_json_api.inspectors.XPropertiesFilter,
            drf_yasg_json_api.inspectors.InlineSerializerInspector,
            drf_yasg_json_api.inspectors.IntegerIDFieldInspector,
            drf_yasg_json_api.inspectors.ManyRelatedFieldInspector,
            drf_yasg.inspectors.RelatedFieldInspector,
            drf_yasg.inspectors.SimpleFieldInspector,
            drf_yasg.inspectors.StringDefaultFieldInspector,
        ]

    class ProjectSerializer(serializers.ModelSerializer):
        class Meta:
            model = test_models.Project
            fields = ('id', 'name', 'archived', 'members')
            extra_kwargs = {
                'name': {'read_only': True},
                'archived': {'read_only': True},
                'members': {'write_only': True},
            }

    class ProjectViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
        queryset = test_models.Project.objects.all()
        serializer_class = ProjectSerializer
        renderer_classes = [renderers.JSONRenderer]
        parser_classes = [parsers.JSONParser]
        swagger_schema = XPropertiesSwaggerAutoSchema

    router = routers.DefaultRouter()
    router.register(r'projects', ProjectViewSet, **compatibility._basename_or_base_name('projects'))

    generator = OpenAPISchemaGenerator(info=openapi.Info(title="", default_version=""), patterns=router.urls)

    swagger = generator.get_schema(request=None, public=True)

    request_body_schema = swagger['paths']['/projects/']['post']['parameters'][0]['schema']['properties']
    # TODO: add support for marking whole data/relationships if all children write only
    # assert 'x-writeOnly' in request_body_schema['data']['properties']['relationships']
    members_schema = request_body_schema['data']['properties']['relationships']['properties']['members']['properties']
    assert 'x-writeOnly' in members_schema['data']['items']['properties']['id']

    response_schema = swagger['paths']['/projects/']['post']['responses']['201']['schema']['properties']
    # TODO: add support for marking whole attributes key read_only if all children read only
    # assert 'x-readOnly' in response_schema['data']['properties']['attributes']
    assert 'readOnly' in response_schema['data']['properties']['attributes']['properties']['name']
    assert 'readOnly' in response_schema['data']['properties']['attributes']['properties']['archived']



import pytest

from django.db import models


class MyModel(models.Model):
    name = models.CharField(max_length=255)


@pytest.mark.django_db
def test_django_models_setup():
    instance = MyModel(name='qwerty')
    instance.save()



# Global test models, define test-specific models locally in test_* files
from django.db import models


class Member(models.Model):
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)


class Project(models.Model):
    name = models.CharField(max_length=100)
    archived = models.BooleanField()
    members = models.ManyToManyField(Member, related_name='projects')
    owner_member = models.ForeignKey(Member, on_delete=models.DO_NOTHING)
    sub_projects = models.ManyToManyField('self')
